// sha1.cc - SHA1
#define VERSION "v1.0"
// GPLed - Stephan T. Lavavej - http://stl.caltech.edu

#include <cstdio>
#include <cstdlib>
using namespace std;

unsigned long int circularLeftShift(unsigned long int ul, unsigned int n);
unsigned long int f(unsigned int t, unsigned long int b, unsigned long int c, unsigned long int d);
FILE * file_open_rb(const char *name);
void file_close(FILE *fp);
unsigned long int readUL(const unsigned char * buf, size_t n);

int main(int argc, char *argv[]) {
    printf("SHA1 " VERSION " [GPLed] by Stephan T. Lavavej, http://stl.caltech.edu" "\n");

    if (argc == 1) {
        printf("GPL:   http://www.gnu.org/licenses/gpl.txt" "\n");
        printf("USAGE: sha1 filename1 [filename2] ...\n");
        printf("       Wildcards such as *.txt are acceptable\n");
        exit(EXIT_SUCCESS);
    }
    for (int z = 1; z < argc; z++) {
        FILE *ifp;
        try {
            ifp = file_open_rb(argv[z]);
        } catch (int) {
            continue;
        }
    
        unsigned long int k[80], w[80], a[5], temp;
        for (int t = 0; t < 20; t++)
            k[t] = 0x5A827999UL;
        for (int t = 20; t < 40; t++)
            k[t] = 0x6ED9EBA1UL;
        for (int t = 40; t < 60; t++)
            k[t] = 0x8F1BBCDCUL;
        for (int t = 60; t < 80; t++)
            k[t] = 0xCA62C1D6UL;
    
        unsigned long int h[5] = {0x67452301UL, 0xEFCDAB89UL, 0x98BADCFEUL, 
                                  0x10325476UL, 0xC3D2E1F0UL};
    
        unsigned long long int filelength = 0;
    
        bool dopadding = false;
        bool keephashing = true;
        bool onewritten = false;
        int charsread = 0;
    
        while (keephashing) {
            unsigned char buf[64];
    
            if (!dopadding) {
                charsread = fread(buf, 1, 64, ifp);
                filelength += charsread;
                if (charsread < 64)
                    dopadding = true;
            }
            if (dopadding) {
                for (int i = charsread; i < 64; i++)
                    buf[i] = 0;
    
                if (charsread > 55) {
                    buf [charsread] = 0x80;
                    onewritten = true;
                    charsread = 0;
                } else {
                    if (!onewritten)
                        buf [charsread] = 0x80;
                    filelength *= 8;
                    for (int i = 0; i < 8; i++)
                        buf[i + 56] = (filelength >> (56 - 8 * i)) & 0xFF;
    
                    keephashing = false;
                }
            }
    
            for (int i = 0; i < 16; i++)
                w[i] = readUL(buf, i * 4);
    
            for (int t = 16; t < 80; t++)
                w[t] = circularLeftShift(w[t - 3] ^ w[t - 8] ^ w[t - 14] ^ w[t - 16], 1);
    
            for (int i = 0; i < 5; i++)
                a[i] = h[i];
    
            for (int t = 0; t < 80; t++) {
                temp = circularLeftShift(a[0], 5) + f(t, a[1], a[2], a[3]) + a[4] + w[t] + k[t];
                a[4] = a[3];
                a[3] = a[2];
                a[2] = circularLeftShift(a[1], 30);
                a[1] = a[0];
                a[0] = temp;
            }
    
            for (int i = 0; i < 5; i++)
                h[i] += a[i];
        }
    
        printf("%08lX %08lX %08lX %08lX %08lX %s\n", h[0], h[1], h[2], h[3], h[4], argv[z]);
        file_close(ifp);
    }
}

unsigned long int f(unsigned int t, unsigned long int b, unsigned long int c, unsigned long int d) {
    if (t <= 19)
        return ((b & c) | (~b & d));
    else if (t >= 40 && t <= 59)
        return ((b & c) | (b & d) | (c & d));
    else return (b ^ c ^ d);
}

FILE * file_open_rb(const char *name) {
    FILE *fp = fopen(name, "rb");
    if (fp == NULL) {
        throw (int) 0;
    }
    return fp;
}

void file_close(FILE *fp) {
    if (fclose(fp) == EOF) {
        printf("ERROR: Problem while closing file.\n");
        exit(EXIT_FAILURE);
    }
}

unsigned long int readUL(const unsigned char * buf, size_t n) {
    unsigned long int ul = 0;
    for (int i = 0; i < 4; i++)
        ul |= (unsigned long int) buf[n + i] << (24 - 8 * i);
    return ul;
}

unsigned long int circularLeftShift(unsigned long int ul, unsigned int n) {
    return (ul << n) | (ul >> (32 - n));
}
